pragma solidity ^0.4.0;

/**
 * Solidity implementation of best tariff matching smart contract variant.
 * Library for global used functions for best tariff matching contracts.
 * license: MIT
 * author: Christoph Leixnering
 * version: 1.3
 * date: 2017-10-26
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 */
library BestTariffLib {
    
    /**
     * Hash value of embedded load forecast.
     * 32 bytes - so use at least sha256 for hashing.
     */
    struct Loadforecasthash {
        bytes32 loadforecasthash;
    }
    
    /**
     * Data structure for hashed template load profiles offered by utilities.
     * @param loadprofilehashes - byte array of hashed template load profiles.
     */
    struct Templatehash {
        bytes32[] loadprofilehashes;
    }
    
    /**
     * Array of offered embedded load profiles of type Templatehash.
     */
    struct Templatehashes {
        mapping (address => Templatehash) templatehashes;
    }
    
    /**
     * Byte array holding a hexadecimal representation of smart meter embedding.
     */
    struct Forecast {
        bytes32[32] forecast;
    }
    
    /**
     * Data structure for calculated distance of utility template load profile.
     * @param utility - address of utility.
     * @param distance - calculated Hamming distance.
     * @param index - index of template load profile.
     */
    struct TemplateDistance {
        address utility;
        uint16 distance;
        uint index;
    }
    
    /**
     * Array of calculated distances of type TemplateDistance.
     */
    struct TemplateDistances {
        TemplateDistance[] templateDistances;
    }
    
    /**
     * Constructor runs only when the contract is created.
     * @param selfLoadforecasthash - storage reference.
     * @param loadforecast - keccak256 hash of load forecast (32 bytes ascii).
     * @return void
     */
    function create(Loadforecasthash storage selfLoadforecasthash, 
        bytes32 loadforecast) public
    {
        selfLoadforecasthash.loadforecasthash = loadforecast;
    }
    
    /**
     * Commit function for utilities to send their embedded load profile hashes.
     * @param selfTemplatehashes - storage reference.
     * @param profiles - array of keccak256 hashed embedded load profiles.
     * @return uint8
     */
    function commit(Templatehashes storage selfTemplatehashes, 
        bytes32[] memory profiles) public returns (uint8 ret)
    {
        // Check for existing load profile hashes.
        if (selfTemplatehashes.templatehashes[msg.sender]
            .loadprofilehashes.length == 0)
        {
            // Load profile hashes received.
            for (uint i = 0; i < profiles.length; i++) {
                selfTemplatehashes.templatehashes[msg.sender]
                    .loadprofilehashes.push(profiles[i]);
            }
            ret = 0;
        } else {
            // Load profile hashes already sent.
            ret = 1;
        }
    }
    
    /**
     * Smart meter preparing load profile forecast commitment for evaluation.
     * @param selfLoadforecasthash - storage reference.
     * @param selfForecast - storage reference.
     * @param owner - address id of contract owner.
     * @param embeddedForecast - bytes32 array of embedded forecast.
     * @param salt - bytes8 of random number used for hash calculation.
     * @return uint8
     */
    function smopen(Loadforecasthash storage selfLoadforecasthash, 
        Forecast storage selfForecast, address owner, 
        bytes32[32] memory embeddedForecast, bytes8 salt) 
        public returns (uint8 ret)
    {
        // Check if message is from contract owner.
        if (msg.sender == owner) {
            // Check hashes and persist load forecast.
            if (getHash(embeddedForecast, salt) == 
                selfLoadforecasthash.loadforecasthash)
            {
                // Load forecast commitment received.
                selfForecast.forecast = embeddedForecast;
                ret = 0;
            } else {
                // Invalid load forecast commitment.
                ret = 2;
            }
        } else {
            // Invalid sender.
            ret = 1;
        }
    }
    
    /**
     * Utility preparing embedded template load profiles for evaluation.
     * @param selfForecast - storage reference.
     * @param selfTemplatehashes - storage reference.
     * @param selfTemplateDistances - storage reference.
     * @param loadprofile - bytes32 array of embedded template load profile.
     * @param salt - bytes8 of random number used for hash calculation.
     * @param map - mapping table for Hamming distance calculation with xor.
     * @return uint8
     */
    function uopen(Forecast storage selfForecast,
        Templatehashes storage selfTemplatehashes,
        TemplateDistances storage selfTemplateDistances,
        bytes32[32] memory loadprofile, bytes8 salt,
        uint8[16] memory map) public
        returns (uint8 ret)
    {
        // Check for existing forecast commitment.
        if (selfForecast.forecast.length > 0) {
            // Check for existing load profile commitment.
            if (selfTemplatehashes.templatehashes[msg.sender]
                .loadprofilehashes.length > 0)
            {
                // Prepare hash for load profile check.
                bytes32 tmploadprofilehash = getHash(loadprofile, salt);
                // Invalid load profile commitment.
                ret = 3;
                for (uint i = 0;
                    i < selfTemplatehashes.templatehashes[msg.sender]
                        .loadprofilehashes.length;
                    i++)
                {
                    // Check hashes.
                    if (tmploadprofilehash == 
                        selfTemplatehashes.templatehashes[msg.sender]
                            .loadprofilehashes[i])
                    {
                        // Calculate distance and store result.
                        uint16 distance = getDistance(
                            selfForecast.forecast,
                            loadprofile,
                            map
                        );
                        // Persist information.
                        TemplateDistance memory templDist;
                        templDist.utility = msg.sender;
                        templDist.distance = distance;
                        templDist.index = i;
                        selfTemplateDistances.templateDistances.push(templDist);
                        // Load profile commitment received.
                        ret = 0;
                        break;
                    }
                }
            } else {
                // Invalid sender, no load profile hashes committed.
                ret = 2;
            }
        } else {
            // Contract not opened for commitment.
            ret = 1;
        }
    }
    
    /**
	 * Find load profile with smallest normalized Hamming distance to forecast.
	 * @param selfTemplateDistances - storage reference.
	 * @param owner - address id of contract owner.
     * @return uint8
     * @return address
     * @return uint
     * @return uint16
     */
    function evaluate(TemplateDistances storage selfTemplateDistances, 
        address owner) public view 
        returns (uint8 ret, address utility, uint index, uint16 distance)
    {
        // Evaluation is only allowed by contract owner.
        if (msg.sender == owner) {
            if (selfTemplateDistances.templateDistances.length > 0) {
                // Get minimum distance - maximum possible is 8192.
                distance = 8193;
                for (uint i = 0; i < selfTemplateDistances.templateDistances
                    .length; i++)
                {
                    if (selfTemplateDistances.templateDistances[i].distance < 
                        distance)
                    {
                        utility = 
                            selfTemplateDistances.templateDistances[i].utility;
                        distance = 
                            selfTemplateDistances.templateDistances[i].distance;
                        index = 
                            selfTemplateDistances.templateDistances[i].index;
                    }
                }
                // Best tariff evaluated.
                ret = 0;
            } else {
                // No utility commitment.
                ret = 2;
            }
        } else {
            // Invalid sender.
            ret = 1;
        }
    }
    
    /**
     * Generates hash of 8 elements of the embedding.
     * @param a - bytes32 of element a.
     * @param b - bytes32 of element b.
     * @param c - bytes32 of element c.
     * @param d - bytes32 of element d.
     * @param e - bytes32 of element e.
     * @param f - bytes32 of element f.
     * @param g - bytes32 of element g.
     * @param h - bytes32 of element h.
     * @return bytes32
     */
    function concat(bytes32 a, bytes32 b, bytes32 c, bytes32 d, bytes32 e,
        bytes32 f, bytes32 g, bytes32 h) internal pure returns (bytes32)
    {
        return keccak256(a, b, c, d, e, f, g, h);
    }
    
    /**
     * Generates hash of embedding and salt.
     * @param e - bytes32[32] array of embedding.
     * @param s - bytes8 salt.
     * @return bytes32
     */
    function getHash(bytes32[32] memory e, bytes8 s) internal pure
        returns (bytes32)
    {
        bytes32[4] memory elems;
        uint8 index = 0;
        for (uint8 i = 0; i < 32; i+=8) {
            elems[index] = concat(e[i], e[i+1], e[i+2],
                e[i+3], e[i+4], e[i+5], e[i+6], e[i+7]);
            index++;
        }
        return keccak256(elems[0], elems[1], elems[2], elems[3], s);
    }
    
    /**
     * Calculates Hamming distance of bytes32 arrays.
     * @param f - customer forecast embedding.
     * @param l - utility load profile embedding.
     * @param map - mapping table for Hamming distance calculation with xor.
     * @return uint16
     */
    function getDistance(bytes32[32] memory f, bytes32[32] memory l, 
        uint8[16] memory map) internal pure returns (uint16)
    {
        uint16 distance = 0;
        for (uint8 i = 0; i < 32; i++) {
            // Bitwise XOR over forecast and load profile.
            bytes32 xor = f[i] ^ l[i];
            for (uint8 ii = 0; ii < 32; ii++) {
                // Calculate distance of left and right 4 bits.
                distance += map[(uint8(xor[ii] & 0xf0)) >> 0x04];
                distance += map[uint8(xor[ii] & 0x0f)];
            }
        }
        return distance;
    }
}
