pragma solidity ^0.4.0;

import "./Lib.sol";

/**
 * Solidity implementation of best tariff matching smart contract variant.
 * license: MIT
 * author: Christoph Leixnering
 * version: 1.3
 * date: 2017-10-26
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING 
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
 * DEALINGS IN THE SOFTWARE.
 */
contract BestTariffVariant {
    
    /**
     * Address id of contract owner.
     * 20 byte value matching to Ethereum address.
     */
    address owner;
    
    /**
     * Mapping table for Hamming distance calculation with xor.
     */
    uint8[16] map = [0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 4];
    
    /**
     * Hash value of embedded load forecast.
     * 32 bytes - so use at least sha256 for hashing.
     */
    BestTariffLib.Loadforecasthash loadforecasthash;
    
    /**
     * Array of offered embedded load profiles of type Templatehash.
     */
    BestTariffLib.Templatehashes templatehashes;
    
    /**
     * Byte array holding a hexadecimal representation of smart meter embedding.
     */
    BestTariffLib.Forecast forecast;
    
    /**
     * Array of calculated distances of type TemplateDistance.
     */
    BestTariffLib.TemplateDistances templateDistances;
    
    /**
     * Event for commit function.
     * @param ret - return code.
     * @return void
     */
    event CommitEvent(uint8 ret);
    
    /**
     * Event for smopen function.
     * @param ret - return code.
     * @return void
     */
    event SmopenEvent(uint8 ret);
    
    /**
     * Event for uopen function.
     * @param ret - return code.
     * @return void
     */
    event UopenEvent(uint8 ret);
    
    /**
     * Event for evaluate function.
     * @param ret - return code.
     * @param utility - address of utility with min distance.
     * @param index - index of load profile of utility with min distance.
     * @param distance - calculated minimum Hamming distance.
     * @return void
     */
    event EvaluateEvent(uint8 ret, address utility, uint index, 
        uint16 distance);
    
    /**
     * Constructor runs only when the contract is created.
     * @param loadforecast - keccak256 hash of load forecast (32 bytes ascii).
     * @return void
     */
    function BestTariffVariant(bytes32 loadforecast) public {
        owner = msg.sender;
        BestTariffLib.create(loadforecasthash, loadforecast);
    }
    
    /**
     * Commit function for utilities to send their embedded load profile hashes.
     * @param profiles - array of keccak256 hashed embedded load profiles.
     * @return uint8
     */
    function commit(bytes32[] memory profiles) public returns (uint8 ret) {
        ret = BestTariffLib.commit(templatehashes, profiles);
        CommitEvent(ret);
    }
    
    /**
     * Smart meter preparing load profile forecast commitment for evaluation.
     * @param embeddedForecast - bytes32 array of embedded forecast.
     * @param salt - bytes8 of random number used for hash calculation.
     * @return uint8
     */
    function smopen(bytes32[32] memory embeddedForecast, bytes8 salt) public
        returns (uint8 ret)
    {
        ret = BestTariffLib.smopen(loadforecasthash, forecast, owner, 
            embeddedForecast, salt);
        SmopenEvent(ret);
    }
    
    /**
     * Utility preparing embedded template load profiles for evaluation.
     * @param loadprofile - bytes32 array of embedded template load profile.
     * @param salt - bytes8 of random number used for hash calculation.
     * @return uint8
     */
    function uopen(bytes32[32] memory loadprofile, bytes8 salt) public
        returns (uint8 ret)
    {
        ret = BestTariffLib.uopen(forecast, templatehashes, templateDistances, 
            loadprofile, salt, map);
        UopenEvent(ret);
    }
    
    /**
	 * Find load profile with smallest normalized Hamming distance to forecast.
     * @return uint8
     * @return address
     * @return uint
     * @return uint16
     */
    function evaluate() public 
        returns (uint8 ret, address utility, uint index, uint16 distance)
    { 
        (ret, utility, index, distance) = 
            BestTariffLib.evaluate(templateDistances, owner);
        EvaluateEvent(ret, utility, index, distance);
    }
}
